package com.huaweicloud.sdk.iot.amqpdemo.android;

import android.view.View;
import android.widget.Toast;

import com.huaweicloud.sdk.iot.amqpdemo.LogUtil;
import com.huaweicloud.sdk.iot.amqpdemo.MainActivity;
import com.huaweicloud.sdk.iot.amqpdemo.client.AmqpClient;
import com.huaweicloud.sdk.iot.amqpdemo.client.AmqpClientOptions;

import org.apache.qpid.proton.amqp.messaging.AmqpValue;
import org.apache.qpid.proton.engine.BaseHandler;
import org.apache.qpid.proton.engine.EndpointState;
import org.apache.qpid.proton.engine.Event;
import org.apache.qpid.proton.engine.Sasl;
import org.apache.qpid.proton.engine.SaslListener;
import org.apache.qpid.proton.engine.Transport;

import java.util.concurrent.atomic.AtomicBoolean;

public class AmqpExample {
    private final MainActivity activity;

    private final AtomicBoolean isStart = new AtomicBoolean();

    private AmqpClient amqpClient;

    private String saslResMsg = "";

    public AmqpExample(MainActivity activity) {
        this.activity = activity;
    }

    public boolean connect() {
        if (isStart.compareAndSet(false, true)) {
            try {
                AmqpClientOptions options = new AmqpClientOptions();
                options.setHost(activity.getConnectAddress().getText().toString());
                options.setPort(Integer.parseInt(activity.getConnectPort().getText().toString()));
                options.setAccessKey(activity.getAccessKey().getText().toString());
                options.setAccessCode(activity.getAccessCode().getText().toString());
                options.setInstanceId(activity.getInstanceId().getText().toString());
                String clientId = activity.getClientId().getText().toString();
                String queueName = activity.getQueueName().getText().toString();
                if (!isBlank(clientId)) {
                    options.setClientId(clientId);
                }
                if (options.checkParamValid() && !isBlank(queueName)) {
                    amqpClient = new AmqpClient(options);
                    amqpClient.newConsumer(queueName, (dlv, msg) -> {
                        String content = ((AmqpValue) msg.getBody()).getValue().toString();
                        LogUtil.info("Receive message: {}", content);
                        activity.getHandler().post(() -> activity.getRecMsgView().setText(content));
                    }, new AmqpLifecycleHandler(), new AmqpSaslListener());
                    return true;
                }
            } catch (Exception e) {
                LogUtil.exceptionError(e, "Amqp example connect error.");
            }
        }
        setMsg("Invalid input params.");
        isStart.set(false);
        return false;
    }

    public void disconnect() {
        if (isStart.compareAndSet(true, false) && amqpClient != null) {
            amqpClient.close();
            onDisConnectView();
            setMsg("Connect closed.");
        }
    }

    public void onConnectView() {
        activity.getHandler().post(() -> {
            activity.getConnectLayout().setVisibility(View.INVISIBLE);
            activity.getDisconnectLayout().setVisibility(View.VISIBLE);
        });
    }

    public void onDisConnectView() {
        activity.getHandler().post(() -> {
            activity.getConnectLayout().setVisibility(View.VISIBLE);
            activity.getDisconnectLayout().setVisibility(View.INVISIBLE);
            activity.getRecMsgView().setText("");
            activity.getConnectButton().setVisibility(View.VISIBLE);
            activity.getStopRetryButton().setVisibility(View.INVISIBLE);
        });
    }

    public void onConnectErrorView() {
        activity.getHandler().post(() -> {
            activity.getConnectLayout().setVisibility(View.VISIBLE);
            activity.getDisconnectLayout().setVisibility(View.INVISIBLE);
            activity.getRecMsgView().setText("");
            activity.getConnectButton().setVisibility(View.INVISIBLE);
            activity.getStopRetryButton().setVisibility(View.VISIBLE);
        });
    }


    public boolean isBlank(String str) {
        return str == null || str.length() == 0;
    }

    private void setMsg(String format, Object... args) {
        String msg = String.format(format.replaceAll("\\{\\}", "%s"), args);
        activity.getHandler().post(() -> Toast.makeText(activity, msg, Toast.LENGTH_LONG).show());
    }

    private class AmqpLifecycleHandler extends BaseHandler {
        @Override
        public void onLinkRemoteOpen(Event e) {
            if (EndpointState.ACTIVE == e.getLink().getRemoteState()) {
                onConnectView();
                setMsg("Connect success.");
            }
        }

        @Override
        public void onTransportError(Event e) {
            onConnectErrorView();
            setMsg("Transport error. msg={}", e.getTransport().getRemoteCondition().toString());
        }

        @Override
        public void onLinkRemoteClose(Event e) {
            onConnectErrorView();
            String additionCause = "";
            if ("amqp:not-allowed".equals(e.getLink().getRemoteCondition().getCondition().toString())) {
                additionCause = "Check whether the queue is exist in IoT platform. ";
            }
            setMsg("Link remote close. {}msg={}", additionCause, e.getLink().getRemoteCondition().toString());
        }

        @Override
        public void onLinkRemoteDetach(Event e) {
            onConnectErrorView();
            setMsg("Link remote detach. msg={}", e.getLink().getRemoteCondition().toString());
        }

        @Override
        public void onSessionRemoteClose(Event e) {
            onConnectErrorView();
            setMsg("Session remote close. msg={}", e.getSession().getRemoteCondition().toString());
        }

        @Override
        public void onConnectionRemoteClose(Event e) {
            onConnectErrorView();
            setMsg("Connection remote close. msg={}", e.getConnection().getRemoteCondition().toString());
        }

        @Override
        public void onConnectionUnbound(Event e) {
            onConnectErrorView();
            if (!isBlank(saslResMsg)) {
                setMsg(saslResMsg);
                saslResMsg = "";
            } else {
                setMsg("Connection unbound, check the network connectivity. msg={}.",
                        e.getConnection().getRemoteCondition().toString());
            }
        }
    }

    private class AmqpSaslListener implements SaslListener {

        @Override
        public void onSaslMechanisms(Sasl sasl, Transport transport) {
        }

        @Override
        public void onSaslInit(Sasl sasl, Transport transport) {
        }

        @Override
        public void onSaslChallenge(Sasl sasl, Transport transport) {
        }

        @Override
        public void onSaslResponse(Sasl sasl, Transport transport) {
        }

        @Override
        public void onSaslOutcome(Sasl sasl, Transport transport) {
            if (Sasl.SaslState.PN_SASL_FAIL == sasl.getState()) {
                LogUtil.error("sasl failed. check accessKey、accessCode and instanceId, sasl={}", sasl);
                saslResMsg = " Sasl failed. Check whether the accessKey、accessCode and instanceId are correct. sasl=" + sasl;
            }
        }
    }
}
