package com.huaweicloud.sdk.iot.amqpdemo.client;


import com.huaweicloud.sdk.iot.amqpdemo.LogUtil;
import com.huaweicloud.sdk.iot.amqpdemo.handler.MessageListener;

import org.apache.qpid.proton.engine.BaseHandler;
import org.apache.qpid.proton.engine.SaslListener;

import java.text.MessageFormat;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

public class AmqpClient {
    private static final int MAX_LINK_COUNT = 10;

    private final AmqpClientOptions options;

    private final AtomicInteger containerIndex = new AtomicInteger(0);

    private final ThreadPoolExecutor threadPoolExecutor;

    private final Map<String, AmqpContainer> amqpContainerMap = new ConcurrentHashMap<>();

    public AmqpClient(AmqpClientOptions options) {
        this.options = options;
        this.threadPoolExecutor = new ThreadPoolExecutor(
                MAX_LINK_COUNT, MAX_LINK_COUNT,
                3L, TimeUnit.MINUTES,
                new LinkedBlockingQueue<>(MAX_LINK_COUNT),
                new ThreadFactory() {
                    private final AtomicLong count = new AtomicLong(0);
                    private final ThreadFactory threadFactory = Executors.defaultThreadFactory();

                    @Override
                    public Thread newThread(Runnable runnable) {
                        Thread thread = threadFactory.newThread(runnable);
                        thread.setName(String.format(Locale.ROOT, "Amqp-" + options.getClientId() + "-%d", count.getAndIncrement()));
                        thread.setUncaughtExceptionHandler((r, e) -> {
                            LogUtil.exceptionError(e, "Thread pool occur uncaught exception. runnable: {}", r);
                        });
                        return thread;
                    }
                },
                (r, e) -> LogUtil.error("Thread pool is busy. execute CustomDiscardPolicy. discard: {}", r));
    }


    public void newConsumer(String queueName, MessageListener messageListener) {
        newConsumer(queueName, messageListener, null, null);
    }

    public synchronized void newConsumer(String queueName, MessageListener messageListener,
                                         BaseHandler lifecycleHandler, SaslListener saslListener) {
        if (threadPoolExecutor.getActiveCount() >= MAX_LINK_COUNT) {
            LogUtil.error("Exceeding max link count [{}].", MAX_LINK_COUNT);
            return;
        }
        if (amqpContainerMap.get(queueName) != null) {
            LogUtil.info("consumer already exist in this client.");
            return;
        }
        String containerId = MessageFormat.format("{0}_{1}", options.getClientId(),
                containerIndex.getAndIncrement());
        AmqpContainer amqpContainer = new AmqpContainer(containerId, options);
        if (lifecycleHandler != null) {
            amqpContainer.addHandler(lifecycleHandler);
        }
        if (saslListener != null) {
            amqpContainer.setSaslListener(saslListener);
        }
        threadPoolExecutor.execute(() -> {
            try {
                amqpContainer.initialize(queueName, messageListener);
                amqpContainer.open();
            } catch (Exception e) {
                LogUtil.exceptionError(e, "create amqp consumer failed.");
            }
        });
        amqpContainerMap.put(queueName, amqpContainer);
    }

    public synchronized void closeConsumer(String queueName) {
        AmqpContainer amqpContainer = amqpContainerMap.get(queueName);
        if (amqpContainer != null) {
            try {
                amqpContainer.close();
                amqpContainerMap.remove(queueName);
            } catch (Exception e) {
                LogUtil.exceptionError(e, "close consumer error");
            }
        }
    }

    public synchronized void close() {
        for (AmqpContainer amqpContainer : amqpContainerMap.values()) {
            try {
                amqpContainer.close();
            } catch (Exception e) {
                LogUtil.exceptionError(e, "close consumer error");
            }
        }
        amqpContainerMap.clear();
        threadPoolExecutor.shutdown();
    }
}
