package com.huaweicloud.sdk.iot.amqpdemo.client;

import com.huaweicloud.sdk.iot.amqpdemo.LogUtil;
import com.huaweicloud.sdk.iot.amqpdemo.handler.AmqpErrorHandler;
import com.huaweicloud.sdk.iot.amqpdemo.handler.ChannelHandler;
import com.huaweicloud.sdk.iot.amqpdemo.handler.ConnectionHandler;
import com.huaweicloud.sdk.iot.amqpdemo.handler.DeliveryHandler;
import com.huaweicloud.sdk.iot.amqpdemo.handler.FlowHandler;
import com.huaweicloud.sdk.iot.amqpdemo.handler.MessageListener;

import org.apache.qpid.proton.amqp.messaging.Source;
import org.apache.qpid.proton.amqp.messaging.Target;
import org.apache.qpid.proton.engine.Collector;
import org.apache.qpid.proton.engine.Connection;
import org.apache.qpid.proton.engine.Event;
import org.apache.qpid.proton.engine.Handler;
import org.apache.qpid.proton.engine.Receiver;
import org.apache.qpid.proton.engine.Sasl;
import org.apache.qpid.proton.engine.SaslListener;
import org.apache.qpid.proton.engine.Session;
import org.apache.qpid.proton.engine.SslDomain;
import org.apache.qpid.proton.engine.Transport;

import java.io.IOException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.text.MessageFormat;
import java.util.List;
import java.util.Timer;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

public class AmqpContainer {
    private final String containerId;

    private final AmqpClientOptions options;

    private Selector selector;

    private Collector collector;

    private Transport transport;

    private Connection connection;

    private Session session;

    private Receiver receiver;

    private Source source;

    private SocketChannel socketChannel;

    private SaslListener saslListener;

    private final Timer heartBeatTimer = new Timer();

    private final AtomicBoolean isInitialize = new AtomicBoolean();

    private final AtomicInteger reconnectCount = new AtomicInteger();

    private final AtomicBoolean isStop = new AtomicBoolean();

    private final List<Handler> handlerList = new CopyOnWriteArrayList<>();

    public AmqpContainer(String containerId, AmqpClientOptions options) {
        this.containerId = containerId;
        this.options = options;
    }

    public String getContainerId() {
        return containerId;
    }

    public AmqpClientOptions getOptions() {
        return options;
    }

    public Selector getSelector() {
        return selector;
    }

    public Transport getTransport() {
        return transport;
    }

    public Connection getConnection() {
        return connection;
    }

    public void addHandler(Handler handler) {
        this.handlerList.add(handler);
    }

    public void setSocketChannel(SocketChannel socketChannel) {
        this.socketChannel = socketChannel;
    }

    public Timer getHeartBeatTimer() {
        return heartBeatTimer;
    }

    public AtomicInteger getReconnectCount() {
        return reconnectCount;
    }

    public void setSaslListener(SaslListener saslListener) {
        this.saslListener = saslListener;
    }

    public void initialize(String queueName, MessageListener messageListener) throws IOException {
        if (isInitialize.compareAndSet(false, true)) {
            this.selector = Selector.open();
            this.collector = Collector.Factory.create();
            this.transport = buildTransport();
            this.source = new Source();
            source.setAddress(queueName);
            initReceiver();
            addHandler(new ConnectionHandler(this));
            addHandler(new FlowHandler(options.getQueuePrefetch()));
            addHandler(new DeliveryHandler(this, messageListener));
            addHandler(new AmqpErrorHandler(queueName, this.containerId));
        }
    }

    private void initReceiver() {
        this.connection = Connection.Factory.create();
        connection.setHostname(options.getHost());
        connection.collect(collector);
        this.session = connection.session();
        this.receiver = session.receiver(MessageFormat.format("{0}_{1}", "link-", this.containerId));
        receiver.setSource(source);
        receiver.setTarget(new Target());
    }

    public void open() {
        if (!isInitialize.get()) {
            LogUtil.error("amqp container has not been initialize.");
            return;
        }
        this.connection.open();
        this.session.open();
        this.receiver.open();

        try {
            while (!isStop.get()) {
                processEvents();
                selector.selectNow();
                if (selector.keys().isEmpty()) {
                    selector.close();
                    return;
                }

                selector.selectedKeys().clear();
                if (selector.select(5000) > 0) {
                    if (!selector.isOpen()) {
                        return;
                    }
                    for (SelectionKey key : selector.selectedKeys()) {
                        ((ChannelHandler) key.attachment()).handler();
                    }
                } else {
                    if (!socketChannel.finishConnect()) {
                        LogUtil.error("connect {} timeout.", options.getEndpointStr());
                        transport.unbind();
                        socketChannel.close();
                    }
                }
            }
        } catch (Exception e) {
            LogUtil.exceptionError(e, "process event error.");
        }
    }

    public void close() {
        if (!isInitialize.get()) {
            LogUtil.error("amqp container has not been initialize.");
            return;
        }
        LogUtil.info("amqp container {} close.", containerId);
        heartBeatTimer.cancel();
        receiver.close();
        session.close();
        connection.close();
        transport.close();
        try {
            isStop.set(true);
            selector.close();
            if (this.socketChannel != null) {
                socketChannel.close();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void rebuild() {
        LogUtil.info("Amqp {} rebuild start.", containerId);
        try {
            receiver.close();
            session.close();
            connection.close();
            transport.close();
            if (this.socketChannel != null) {
                socketChannel.close();
            }
            transport = buildTransport();
            initReceiver();
            connection.open();
            session.open();
            receiver.open();
        } catch (Exception e) {
            LogUtil.exceptionError(e, "Amqp rebuild failed.");
        }
    }

    private void processEvents() {
        while (!isStop.get()) {
            Event ev = collector.peek();
            if (ev == null) {
                break;
            }
//             LogUtil.info("eventType: {}", ev.getType());
            for (Handler h : handlerList) {
                ev.dispatch(h);
            }
            collector.pop();
        }
    }

    public Transport buildTransport() {
        Transport transport = Transport.Factory.create();
        Sasl sasl = transport.sasl();
        if (saslListener != null) {
            sasl.setListener(saslListener);
        }
        sasl.plain(generateUserName(), options.getAccessCode());
        sasl.client();
        SslDomain clientSslDomain = SslDomain.Factory.create();
        clientSslDomain.init(SslDomain.Mode.CLIENT);
        clientSslDomain.setPeerAuthentication(SslDomain.VerifyMode.ANONYMOUS_PEER);
        transport.ssl(clientSslDomain);
        transport.setIdleTimeout(options.getIdleTimeout());
        return transport;
    }

    private String generateUserName() {
        String userName = "accessKey=" + options.getAccessKey() + "|timestamp=" + System.currentTimeMillis();
        String instanceId = options.getInstanceId();
        if (instanceId != null && instanceId.trim().length() > 0) {
            userName += "|instanceId=" + instanceId;
        }
        return userName;
    }
}
