package com.huaweicloud.sdk.iot.amqpdemo.handler;


import com.huaweicloud.sdk.iot.amqpdemo.LogUtil;
import com.huaweicloud.sdk.iot.amqpdemo.client.AmqpClientOptions;
import com.huaweicloud.sdk.iot.amqpdemo.client.AmqpContainer;

import org.apache.qpid.proton.engine.BaseHandler;
import org.apache.qpid.proton.engine.Connection;
import org.apache.qpid.proton.engine.EndpointState;
import org.apache.qpid.proton.engine.Event;
import org.apache.qpid.proton.engine.Transport;

import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.security.SecureRandom;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class ConnectionHandler extends BaseHandler {
    private static final long DEFAULT_BACKOFF = 1000L;

    private final AmqpContainer container;

    private final Timer heartBeatTimer;

    private final AtomicInteger reconnectCount;

    private final SecureRandom random = new SecureRandom();

    public ConnectionHandler(AmqpContainer container) {
        this.container = container;
        this.heartBeatTimer = container.getHeartBeatTimer();
        this.reconnectCount = container.getReconnectCount();
    }

    @Override
    public void onConnectionLocalOpen(Event evt) {
        Transport transport = container.getTransport();
        Connection conn = container.getConnection();
        if (conn.getRemoteState() == EndpointState.UNINITIALIZED) {
            conn.setContainer(container.getContainerId());
            transport.bind(conn);
            try {
                SocketChannel socket = SocketChannel.open();
                ChannelHandler channelHandler = new ChannelHandler(container);
                transport.setContext(channelHandler);
                socket.configureBlocking(false);
                SelectionKey selectionKey = socket.register(container.getSelector(), SelectionKey.OP_CONNECT,
                        channelHandler);
                channelHandler.init(socket, selectionKey);
                container.setSocketChannel(socket);
                LogUtil.info("Begin connect: {}", container.getOptions().getEndpointStr());
                socket.connect(
                        new InetSocketAddress(container.getOptions().getHost(), container.getOptions().getPort()));
            } catch (Exception e) {
                LogUtil.exceptionError(e, "socket connect failed.");
            }
        }
    }

    @Override
    public void onConnectionRemoteOpen(Event e) {
        reconnectCount.set(0);
        executeHeartBeatCheck();
    }

    @Override
    public void onLinkRemoteClose(Event e) {
        reconnect();
    }

    @Override
    public void onTransport(Event evt) {
        ((ChannelHandler) evt.getTransport().getContext()).handler();
    }

    @Override
    public void onConnectionRemoteClose(Event evt) {
        reconnect();
    }

    @Override
    public void onConnectionUnbound(Event evt) {
        reconnect();
    }

    private void reconnect() {
        AmqpClientOptions options = container.getOptions();
        long maxReconnectAttempts = options.getMaxReconnectAttempts();
        if (maxReconnectAttempts > 0 && reconnectCount.get() > maxReconnectAttempts) {
            LogUtil.error("Connect: {} already exceeding max reconnect attempts {}.", options.getEndpointStr(),
                    maxReconnectAttempts);
            return;
        }
        long reconnectDelay = options.getReconnectDelay();
        int lowBound = (int) (DEFAULT_BACKOFF * 0.8);
        int highBound = (int) (DEFAULT_BACKOFF * 1.0);
        long randomBackOff = random.nextInt(highBound - lowBound);
        int powParameter = reconnectCount.getAndIncrement() & 0x0F;
        long backOffWithJitter = (long) (Math.pow(2.0, powParameter)) * (randomBackOff + lowBound);
        long waitTimeUntilNextRetry = Math.min(reconnectDelay + backOffWithJitter, options.getMaxReconnectDelay());
        try {
            LogUtil.info("Amqp reconnect wait time: {}", waitTimeUntilNextRetry);
            Thread.sleep(waitTimeUntilNextRetry);
        } catch (InterruptedException e) {
            LogUtil.exceptionError(e, "Amqp reconnect wait failed. Execute immediately");
        }
        container.rebuild();
    }

    private void executeHeartBeatCheck() {
        long now = TimeUnit.NANOSECONDS.toMillis(System.nanoTime());
        long deadline = container.getTransport().tick(now);
        if (deadline != 0) {
            heartBeatTimer.schedule(new HeartBeatCheck(), Math.max(deadline - now, 1));
        }
    }

    private final class HeartBeatCheck extends TimerTask {
        @Override
        public void run() {
            if (container.getConnection().getLocalState() == EndpointState.ACTIVE) {
                if (container.getTransport().isClosed()) {
                    LogUtil.error("HeartBeatCheck closed by server exceeding client idle timeout.");
                } else {
                    executeHeartBeatCheck();
                }
            } else {
                LogUtil.error("HeartBeatCheck closed, connection is not active.");
            }
        }
    }
}
